#ifndef BMP_HEADERS_H
#define BMP_HEADERS_H

#include <iosfwd>
#include <cstdint>

//
// Le format des bitmaps a t pris sur
// http://en.wikipedia.org/wiki/Windows_bitmap
//
template <int SZ>
   struct base_raw_header {
      static constexpr auto TAILLE = SZ;
      std::uint8_t raw_data[TAILLE];
   protected:
      size_t make_size_t(int offs) const noexcept {
         return size_t(raw_data[offs] | (raw_data[offs + 1] << 8) | (raw_data[offs + 2] << 16) | (raw_data[offs + 3] << 24) );
      }
      std::int32_t make_int32(int offs) const noexcept {
         return static_cast<std::int32_t>(raw_data[offs] | (raw_data[offs + 1] << 8) | (raw_data[offs + 2] << 16) | (raw_data[offs + 3] << 24) );
      }
      std::int16_t make_int16(int offs) const noexcept {
         return static_cast<std::int16_t>(raw_data[offs] | (raw_data[offs + 1] << 8));
      }
      friend std::istream& operator>>(std::istream &is, base_raw_header &hdr) {
         if (!is) return is;
         char val;
         for (int i = 0; i < base_raw_header::TAILLE && is.get(val); ++i)
            hdr.raw_data[i] = val;
         return is;
      }
   };

struct bitmap_header : base_raw_header<14> {
   static const std::uint8_t MAGIC_NUMBER[];
   bool is_valid() const noexcept {
      return raw_data[0] == MAGIC_NUMBER[0] && raw_data[1] == MAGIC_NUMBER[1];
   }
   size_t size() const noexcept {
      return make_size_t(2);
   }
   size_t offset() const noexcept {
      return make_size_t(10);
   }
};

struct windows_v3_header : base_raw_header<40> {
   enum CompressionMethod {
      CMETH_BI_RGB, CMETH_BI_RLE8, CMETH_BI_RLE4, CMETH_BI_BITFIELDS, CMETH_BI_JPEG, CMETH_BI_PNG
   };
   bool is_valid() const noexcept {
      return size() == TAILLE;
   }
   size_t size() const noexcept {
      return make_size_t(0);
   }
   size_t width() const noexcept {
      return make_size_t(4);
   }
   size_t height() const noexcept {
      return make_size_t(8);
   }
   std::int16_t nb_color_panes() const noexcept {
      return make_int16(12);
   }
   std::int16_t color_depth() const noexcept {
      return make_int16(14);
   }
   CompressionMethod compression_method() const noexcept {
      return static_cast<CompressionMethod>(make_size_t(16));
   }
   size_t raw_size() const noexcept {
      return make_size_t(20);
   }
   std::int32_t horizontal_resolution() const noexcept {
      return make_int32(24);
   }
   std::int32_t vertical_resolution() const noexcept {
      return make_int32(28);
   }
   std::int32_t colors_in_palette() const noexcept {
      return make_int32(32);
   }
   std::int32_t important_colors() const noexcept {
      return make_int32(36);
   }
};

std::ostream& operator<<(std::ostream&, const windows_v3_header::CompressionMethod);
std::ostream& operator<<(std::ostream&, const windows_v3_header&);

#include <vector>
struct bitmap {
   bitmap_header hdr;
   windows_v3_header wv3hdr;
   std::vector<std::uint8_t> raw_data;
};

std::istream& operator>>(std::istream&, bitmap&);

#endif
